/*
** Module   : BufferFieldImpl.java
** Abstract : Implementation of the BufferField interface
**
** Copyright (c) 2013-2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description---------------------------------
** 001 ECF 20150315 Created initial version by refactoring all support specific to BUFFER-FIELD
**                  4GL compatibility from the FieldReference class. See that class for a full
**                  history. This refactoring was done to eliminate the memory footprint and
**                  overhead of the BUFFER-FIELD implementation from the many simpler use cases
**                  for the FieldReference class.
** 002 ECF 20150320 Javadoc fix.
** 003 OM  20150415 Added LITERAL-QUESTION attribute implementation.
** 004 SVL 20150503 Added bufferHandle.
*/

package com.goldencode.p2j.persist;

import java.lang.reflect.*;
import java.util.logging.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.p2j.util.UniqueIdGenerator.IdKind;

/**
 * Implementation of Progress 4GL BUFFER-FIELD methods and attributes, using a {@link
 * FieldReference} as a delegate worker.
 */
public final class BufferFieldImpl
extends HandleChain
implements BufferField,
           Deletable
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(BufferField.class.getName());
   
   /** {@link FieldReference} delegate object */
   private final FieldReference fieldRef;
   
   /** Corresponds to ADM-DATA attribute. */
   private final character admData = new character();
   
   /** Corresponds to UNIQUE-ID attribute. */
   private final integer uniqueID =
      new integer(UniqueIdGenerator.getUniqueId(IdKind.BUFFER_FIELD));
   
   /**
    * Create a new field reference given a DMO proxy and property name.  A non-indexed getter
    * method will be used to resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    */
   public BufferFieldImpl(DataModelObject dmo, String property)
   {
      this(dmo, property, false, null);
   }
   
   /**
    * Create a new field reference given a DMO proxy, a property name, and
    * an index for the property. An indexed getter method will be used to
    * resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    * @param   uppercase
    *          <code>true</code> to force a <code>character</code> value to be
    *          uppercased/right-trimmed;
    *          <code>false</code> to leave the result unchanged.
    * @param   index
    *          Index to pass to property's getter method at resolve time.
    *
    * @throws  IllegalArgumentException
    *          if field reference is not recognized.
    */
   private BufferFieldImpl(DataModelObject dmo,
                           String property,
                           boolean uppercase,
                           NumberType index)
   {
      fieldRef = new FieldReference(dmo, property, uppercase, index);
   }
   
   /**
    * Get the <code>name</code> attribute of a BUFFER-FIELD, which is the legacy name of the
    * field referenced by this object.
    * 
    * @return  See above.
    */
   @Override
   public character name()
   {
      if (name.isUnknown())
      {
         // assign directly to the variable, rather than through the API, to avoid a 4052 error
         String property = fieldRef.getPropertyName();
         Object dmo = fieldRef.getDMO();
         this.name.assign(TableMapper.getLegacyFieldName((Buffer) dmo, property));
      }
      
      return super.name();
   }
   
   /**
    * Conversion of EXTENT attribute (KW_EXTENT).
    *
    * Getter of EXTENT attribute
    *
    */
   public integer extent()
   {
      Method sizer = fieldRef.getSizer();
      
      if (sizer == null)
      {
         return new integer(0);
      }
      
      try
      {
         Object dmo = fieldRef.getDMO();
         
         return (integer) sizer.invoke(dmo);
      }
      catch (IllegalAccessException | InvocationTargetException e)
      {
         throw new RuntimeException(e);
      }
   }
   
   /**
    * Conversion of INITIAL attribute (KW_INIT).
    *
    * Getter of INITIAL attribute.
    *
    * The value of the INITIAL schema field (which is always CHARACTER), formatted with the
    * buffer-field's format. If the INITIAL schema field has the Unknown value (?), the
    * value of the INITIAL attribute is the null string ("").
    *
    * @return  Current value of INITIAL attribute.
    */
   public character initial()
   {
      String property = fieldRef.getPropertyName();
      String init = TableMapper.getLegacyFieldInitial(getDMOProxy(), property);
      character initial = character.valueOf(init);
      Class<?> cls = fieldRef.getType();
      BaseDataType baseDataType;
      
      try
      {
         Class[] parmTypes = { character.class };
         Constructor ctor = cls.getConstructor(parmTypes);
         baseDataType = (BaseDataType) ctor.newInstance(initial);
      }
      catch (Exception exc)
      {
         throw new RuntimeException(exc);
      }
      
      return character.valueOf(baseDataType.toString(getFormat().getValue()));
   }
   
   /**
    * Conversion of STRING-VALUE attribute (KW_STR_VAL).
    *
    * Getter of STRING-VALUE attribute.
    *
    * The STRING-VALUE( ) method uses the format attribute to convert the buffer value to
    * a string.
    *
    * @return  current value of STRING-VALUE attribute.
    */
   public character stringValue()
   {
      Method sizer = fieldRef.getSizer();
      if (sizer == null)
      {
         return character.valueOf(value().toString(getFormat().getValue()));
      }
      else
      {
         ErrorManager.recordOrShowError(26, "Array subscript 0 is out of range", false);
         
         String msg = "Unable to extract BUFFER-VALUE for field %s";
         String property = fieldRef.getPropertyName();
         String err = String.format(msg, property);
         ErrorManager.recordOrShowError(7366, err, false);
         
         return new character();
      }
   }
   
   /**
    * Conversion of STRING-VALUE attribute (KW_STR_VAL).
    *
    * Getter of STRING-VALUE attribute.
    *
    * The STRING-VALUE( ) method uses the format attribute to convert the buffer value to
    * a string.
    *
    * @param   index
    *          An integer expression representing a subscript, for fields that
    *          have extents.
    *
    * @return  current value of STRING-VALUE attribute.
    */
   public character stringValue(int index)
   {
      return stringValue(new integer(index));
   }
   
   /**
    * Conversion of STRING-VALUE attribute (KW_STR_VAL).
    *
    * Getter of STRING-VALUE attribute.
    *
    * The STRING-VALUE( ) method uses the format attribute to convert the buffer value to
    * a string.
    *
    * @param    index
    *           An integer expression representing a subscript, for fields that
    *           have extents.
    *
    * @return   Current value of STRING-VALUE attribute.
    */
   public character stringValue(NumberType index)
   {
      return character.valueOf(value(index).toString(getFormat().getValue()));
   }
   
   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Getter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the
    * new value.
    *
    * @param   index
    *          An INTEGER expression representing a subscript, for fields
    *          that have extents.
    *
    * @return  Current value of a buffer-field object.
    */
   public BaseDataType value(int index)
   {
      return value(new integer(index));
   }
   
   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Getter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the new value.
    *
    * @param   index
    *          An INTEGER expression representing a subscript, for fields
    *          that have extents.
    *
    * @return  Current value of a buffer-field object if the object is an extent field or
    *          <code>unknown</code> if it is a scalar one.
    */
   public BaseDataType value(NumberType index)
   {
      if (!validateBufferValueIndex(index))
      {
         return fieldRef.unknownValue();
      }
      
      return (index.intValue() == 0 ? value() : getElementReference(index).value());
   }
   
   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Setter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the
    * new value.
    *
    * @param   index
    *          An INTEGER expression representing a subscript, for fields
    *          that have extents.
    * @param   newValue
    *          New value of a buffer-field object.
    *
    */
   public void changeValue(int index, Object newValue)
   {
      changeValue(new integer(index), newValue);
   }
   
   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Setter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the
    * new value.
    *
    * @param   index
    *          An INTEGER expression representing a subscript, for fields
    *          that have extents.
    * @param   newValue
    *          New value of a buffer-field object.
    *
    */
   public void changeValue(NumberType index, Object newValue)
   {
      if (!this.validateBufferValueIndex(index))
      {
         return;
      }
      
      if (index.intValue() == 0)
      {
         changeValue(newValue);
      }
      else
      {
         getElementReference(index).changeValue(newValue);
      }
   }
   
   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Getter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the new value.
    *
    * @return  Current value of a buffer-field object. If the field is an extent field,
    *          <code>unknown</code> value is returned.
    */
   public BaseDataType value()
   {
      return fieldRef.getSizer() == null ? fieldRef.get() : fieldRef.unknownValue();
   }
   
   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    * <p>
    * Setter of current value of a buffer-field object. If you modify the BUFFER-VALUE attribute,
    * Progress sets the buffer-field object to the new value.
    * <p> 
    * Special care is taken when the field is a CHARACTER and the LITERAL-QUESTION is the default
    * value {@code false}: in this case the "?" is understand as the unknown literal (?) and the
    * other strings are rtrimmed. If LITERAL-QUESTION is set to {@code yes} the string are not
    * affected.
    *
    * @param   newValue
    *          New value of a buffer-field object.
    *
    */
   public void changeValue(Object newValue)
   {
      BaseDataType bdt = null;
      
      if (!(newValue instanceof BaseDataType))
      {
         // try some simple conversions
         if (newValue instanceof Integer)
         {
            bdt = new integer((Integer) newValue);
         }
         else if (newValue instanceof Long)
         {
            bdt = new integer((Long) newValue);
         }
         else if (newValue instanceof Double)
         {
            bdt = new decimal((Double) newValue);
         }
         else if (newValue instanceof Boolean)
         {
            bdt = new logical((Boolean) newValue);
         }
         else if (newValue instanceof String)
         {
            bdt = new character((String) newValue);
         }
         else
         {
            throw new ClassCastException("Can't convert " +
                                         newValue.getClass().getName() +
                                         " to BaseDataType!");
         }
      }
      else if (newValue instanceof unknown)
      {
         bdt = fieldRef.unknownValue();
      }
      else if (!fieldRef.getType().isAssignableFrom(newValue.getClass()))
      {
         try
         {
            // try to invoke a polymorphic constructor of the field's type
            Constructor<?> ctor = fieldRef.getType().getConstructor(BaseDataType.class);
            bdt = (BaseDataType) ctor.newInstance(newValue);
         }
         catch (InvocationTargetException exc)
         {
            // the constructor could not manage to convert the passed value into a meaningful
            // object of the field's type
            Throwable cause = exc.getCause();
            if (cause instanceof ConditionException)
            {
               throw (ConditionException) cause;
            }
            
            LOG.log(Level.SEVERE, "Error performing runtime type conversion", exc);
            
            return;
         }
         catch (NoSuchMethodException  |
                InstantiationException |
                IllegalAccessException |
                IllegalArgumentException exc)
         {
            ErrorManager.recordOrThrowError(
               5729,
               "Incompatible datatypes found during runtime conversion");
            return;
         }
      }
      else
      {
         // finally... the simple case
         bdt = (BaseDataType) newValue;
         
         // special handling for character fields when LITERAL-QUESTION attribute is false
         if (bdt instanceof character              &&
             fieldRef.getType() == character.class &&
             !literalQuestion().booleanValue())
         {
            character newChar = ((character) bdt);
            if ("?".equals(newChar.toStringMessage()))
            {
               bdt = newChar.instantiateUnknown();
            }
            else
            {
               bdt = TextOps.rightTrim(newChar);
            }
         }
      }
      
      fieldRef.set(bdt);
   }
   
   /**
    * Conversion of COLUMN-LABEL attribute (KW_COL_LAB).
    *
    * Getter of COLUMN-LABEL attribute.
    *
    * @return  current value of COLUMN-LABEL attribute.
    */
   public character columnLabel()
   {
      String property = fieldRef.getPropertyName();
      
      return character.valueOf(TableMapper.getLegacyFieldColumnLabel(getDMOProxy(), property));
   }
   
   /**
    * Conversion of COLUMN-LABEL attribute (KW_COL_LAB).
    *
    * Getter of COLUMN-LABEL attribute.
    *
    * @param   columnLabel
    *          New value of a attribute.
    */
   public void changeColumnLabel(character columnLabel)
   {
      setColumnLabel(columnLabel);
   }
   
   /**
    * Conversion of BUFFER-NAME attribute (KW_BUF_NAME).
    *
    * Gets the name of the buffer object to which the buffer-field object belongs.
    *
    * @return  name of the buffer object to which the buffer-field object belongs.
    */
   public character bufferName()
   {
      return character.valueOf(fieldRef.getParentBuffer().getDMOAlias());
   }

   /**
    * Returns the handle of the buffer object to which the buffer-field belongs.
    *
    * @return the handle of the buffer object to which the buffer-field belongs.
    */
   public handle bufferHandle()
   {
      return new handle((Buffer) fieldRef.getParentBuffer().getDMOProxy());
   }
   
   /**
    * Conversion of LITERAL-QUESTION attribute (KW_LIT_QSTN).
    *
    * Getter of LITERAL-QUESTION attribute.
    *
    * @return  current value of LITERAL-QUESTION attribute.
    */
   public logical literalQuestion()
   {
      String property = fieldRef.getPropertyName();
      
      return new logical(TableMapper.getLegacyFieldLiteralQuestion(getDMOProxy(), property));
   }
   
   /**
    * Conversion of LITERAL-QUESTION attribute (KW_LIT_QSTN).
    *
    * Setter of LITERAL-QUESTION attribute.
    *
    * @param   literalQuestion
    *          New value of LITERAL-QUESTION attribute.
    */
   public void changeLiteralQuestion(logical literalQuestion)
   {
      String property = fieldRef.getPropertyName();
      TableMapper.setLegacyFieldLiteralQuestion(getDMOProxy(), property, literalQuestion);
   }
   
   /**
    * Conversion of VALIDATE-MESSAGE attribute (KW_VAL_MSG).
    *
    * Getter of VALIDATE-MESSAGE attribute.
    *
    * @return  current value of VALIDATE-MESSAGE attribute.
    */
   public character validateMessage()
   {
      String property = fieldRef.getPropertyName();
      
      return new character(TableMapper.getLegacyFieldValidateMessage(getDMOProxy(), property));
   }
   
   /**
    * Conversion of VALIDATE-MESSAGE attribute (KW_VAL_MSG).
    *
    * Setter of VALIDATE-MESSAGE attribute.
    *
    * @param   validateMessage
    *          New value of VALIDATE-MESSAGE attribute.
    */
   public void changeValidateMessage(character validateMessage)
   {
      String property = fieldRef.getPropertyName();
      TableMapper.setLegacyFieldValidateMessage(getDMOProxy(),
                                                property,
                                                validateMessage.getValue());
   }
   
   /**
    * Conversion of MANDATORY attribute (KW_MAND).
    *
    * Getter of KW_MAND attribute.
    *
    * @return current value of KW_MAND attribute.
    */
   public logical mandatory()
   {
      String property = fieldRef.getPropertyName();
      
      return new logical(TableMapper.isLegacyFieldMandatory(getDMOProxy(), property));
   }
   
   /**
    * Conversion of VALIDATE-EXPRESSION attribute (KW_VAL_EXPR).
    *
    * Getter of VALIDATE-EXPRESSION attribute.
    *
    * @return current value of VALIDATE-EXPRESSION attribute.
    */
   public character validateExpression()
   {
      String property = fieldRef.getPropertyName();
      
      return new character(TableMapper.getLegacyFieldValidateExpression(getDMOProxy(), property));
   }
   
   /**
    * Conversion of VALIDATE-EXPRESSION attribute (KW_VAL_EXPR).
    *
    * Setter of VALIDATE-EXPRESSION attribute.
    *
    * @param  validateExpression
    *         New value of VALIDATE-EXPRESSION attribute.
    */
   public void changeValidateExpression(character validateExpression)
   {
      String property = fieldRef.getPropertyName();
      TableMapper.setLegacyFieldValidateExpression(getDMOProxy(),
                                                   property,
                                                   validateExpression.getValue());
   }
   
   /**
    * Reports if this object is valid for use.
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      return getDMOProxy().valid();
   }
   
   /**
    * Conversion of POSITION attribute (KW_POS).
    *
    * Getter of POSITION attribute.
    *
    * @return current value of POSITION attribute.
    */
   public integer position()
   {
      String property = fieldRef.getPropertyName();
      integer position = TableMapper.getLegacyFieldId(getDMOProxy(), property);
      position.increment();
      
      return position;
   }
   
   /**
    * Returns the state of the DBNAME attribute
    *
    * @return    Current value of the DBNAME attribute.
    */
   public character getDbName()
   {
      return new character(fieldRef.getParentBuffer().getDatabase().getName());
   }
   
   /**
    * Returns the state of the TABLE attribute
    *
    * @return    Current value of the TABLE attribute.
    */
   public character getTable()
   {
      return new character(fieldRef.getParentBuffer().getTable());
   }
   
   /**
    * HEIGHT-CHARS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #invalidAttribute} to display appropriate error.
    * 
    * @return  unknown value.
    */
   public decimal getHeightChars()
   {
      invalidAttribute("HEIGHT-CHARS");
      
      return new decimal();
   }
   
   /**
    * HEIGHT-CHARS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    * 
    * @param   h
    *          Not used.
    */
   public void setHeightChars(NumberType h)
   {
      readOnlyError("HEIGHT");
   }
   
   /**
    * HEIGHT-CHARS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    * 
    * @param   h
    *          Not used.
    */
   public void setHeightChars(double h)
   {
      readOnlyError("HEIGHT");
   }
   
   /**
    * Get the WIDTH-CHARS attribute of this BUFFER-FIELD handle.
    *
    * For buffer-field objects, the WIDTH-CHARS attribute is the number of characters in the
    * STRING-VALUE, which the AVM calculates using the FORMAT attribute.
    *
    * @return  See above.
    */
   public decimal getWidthChars()
   {
      String strValue = stringValue().toStringMessage();
      
      return new decimal(strValue == null ? 0 : strValue.length());
   }
   
   /**
    * Set the WIDTH-CHARS attribute of this BUFFER-FIELD handle.
    *
    * The WIDTH-CHARS attribute of a buffer-field is readable but not writeable.
    *
    * @param   widthChars
    *          The new value for the WIDTH-CHARS attribute.
    */
   public void setWidthChars(NumberType widthChars)
   {
      readOnlyError("WIDTH");
   }
   
   /**
    * Set the WIDTH-CHARS attribute of this BUFFER-FIELD handle.
    *
    * The WIDTH-CHARS attribute of a buffer-field is readable but not writeable.
    * 
    * @param   widthChars
    *          The new value for the WIDTH-CHARS attribute.
    */
   public void setWidthChars(double widthChars)
   {
      readOnlyError("WIDTH");
   }
   
   /**
    * WIDTH-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #invalidAttribute} to display appropriate error.
    * 
    * @return  unknown value.
    */
   public integer getWidthPixels()
   {
      invalidAttribute("WIDTH-PIXELS");
      
      return new integer();
   }
   
   /**
    * WIDTH-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    *
    * @param    widthPixels
    *           Not used.
    */
   public void setWidthPixels(NumberType widthPixels)
   {
      readOnlyError("WIDTH-PIXELS");
   }
   
   /**
    * WIDTH-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    *
    * @param    widthPixels
    *           Not used.
    */
   public void setWidthPixels(int widthPixels)
   {
      readOnlyError("WIDTH-PIXELS");
   }
   
   /**
    * HEIGHT-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #invalidAttribute} to display appropriate error.
    * 
    * @return  unknown value.
    */
   public integer getHeightPixels()
   {
      invalidAttribute("HEIGHT-PIXELS");
      
      return new integer();
   }
   
   /**
    * HEIGHT-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    *
    * @param    heightPixels
    *           Not used.
    */
   public void setHeightPixels(NumberType heightPixels)
   {
      readOnlyError("HEIGHT-PIXELS");
   }
   
   /**
    * HEIGHT-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    *
    * @param    heightPixels
    *           Not used.
    */
   public void setHeightPixels(int heightPixels)
   {
      readOnlyError("HEIGHT-PIXELS");
   }
   
   /**
    * Getter of FORMAT attribute.
    *
    * @return   current value of FORMAT attribute.
    */
   @Override
   public character getFormat()
   {
      String property = fieldRef.getPropertyName();
      return character.valueOf(TableMapper.getLegacyFieldFormat(getDMOProxy(), property));
   }
   
   /**
    * Setter of FORMAt attribute.
    *
    * @param   format
    *          The new value.
    */
   @Override
   public void setFormat(String format)
   {
      setFormat(new character(format));
   }
   
   /**
    * Setter of FORMAt attribute.
    *
    * @param   format
    *          The new value.
    */
   @Override
   public void setFormat(character format)
   {
      String property = fieldRef.getPropertyName();
      TableMapper.setLegacyFieldFormat(getDMOProxy(), property, format.getValue());
   }
   
   /**
    * Getter of COLUMN-LABEL attribute.
    *
    * @return   current value of COLUMN-LABEL attribute.
    */
   public character getColumnLabel()
   {
      return columnLabel();
   }
   
   /**
    * Setter of COLUMN-LABEL attribute.
    * 
    * @param    columnLabel
    *           New value of a attribute.
    */
   public void setColumnLabel(character columnLabel)
   {
      setColumnLabel(columnLabel.getValue());
   }
   
   /**
    * Setter of COLUMN-LABEL attribute.
    * 
    * @param    columnLabel
    *           New value of a attribute.
    */
   public void setColumnLabel(String columnLabel)
   {
      String property = fieldRef.getPropertyName();
      TableMapper.setLegacyFieldColumnLabel(getDMOProxy(), property, columnLabel);
   }
   
   /**
    * Returns the value of the DATA-TYPE attribute
    *
    * @return   Current value of the DATA-TYPE attribute.
    */
   public character getDataType()
   {
      return character.valueOf(fieldRef.getType().getSimpleName());
   }
   
   /**
    * Set the <code>DATA-TYPE</code> attribute..
    *
    * @param    dataType
    *           The DATA-TYPE to be set.
    */
   public void setDataType(String dataType)
   {
      readOnlyError("DATA-TYPE");
   }
   
   /**
    * Set the <code>DATA-TYPE</code> attribute..
    *
    * @param    dataType
    *           The DATA-TYPE to be set.
    */
   public void setDataType(character dataType)
   {
      readOnlyError("DATA-TYPE");
   }
   
   /**
    * Returns the value of the LABEL attribute
    *
    * @return    Current value of the LABEL attribute.
    */
   public character getLabel()
   {
      String property = fieldRef.getPropertyName();
      
      return character.valueOf(TableMapper.getLegacyFieldLabel(getDMOProxy(), property));
   }
   
   /**
    * Set the <code>LABEL</code> attribute of handle.
    *
    * @param    label
    *           The LABEL to be set.
    */
   public void setLabel(String label)
   {
      String property = fieldRef.getPropertyName();
      TableMapper.setLegacyFieldLabel(getDMOProxy(), property, label);
   }
   
   /**
    * Set the <code>LABEL</code> attribute of handle.
    *
    * @param    label
    *           The LABEL to be set.
    */
   public void setLabel(character label)
   {
      setLabel(label.getValue());
   }

   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getADMData()
   {
      return new character(this.admData);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(String value)
   {
      this.admData.assign(value);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(character value)
   {
      this.admData.assign(value);
   }

   /**
    * Gets the the unique ID number associated to this object by the underlying system. Not the 
    * same as the handle value itself.
    * 
    * @return   The integer number of the children.
    */
   @Override
   public integer getUniqueID()
   {
      return new integer(this.uniqueID);
   }
   
   /**
    * Delete the resource.
    *
    * BUFFER-FIELD will be deleted only when the dynamic buffer is deleted. No resource-specific
    * delete is required.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      return false;
   }
   
   /**
    * Check if this resource supports the PREV-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasPrevSibling()
   {
      return false;
   }
   
   /**
    * Check if this resource supports the NEXT-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasNextSibling()
   {
      return false;
   }
   
   /**
    * Get the field reference for the given element of this extent field.
    *
    * @param  index
    *         1-based index of the extent field.
    *
    * @return field reference for the specific extent element.
    */
   private BufferFieldImpl getElementReference(NumberType index)
   {
      NumberType idx = MathOps.minus(index, 1);
      Object dmo = fieldRef.getDMO();
      String property = fieldRef.getPropertyName();
      boolean uppercase = fieldRef.isUpperCase();
      
      return new BufferFieldImpl((DataModelObject) dmo, property, uppercase, idx);
   }
   
   /**
    * Get the DMO proxy which references parent buffer.
    *
    * @return  DMO proxy object.
    */
   private Buffer getDMOProxy()
   {
      return (Buffer) fieldRef.getParentBuffer().getDMOProxy();
   }
   
   /**
    * Validate the index passed to a BUFFER-VALUE method call. For scalar fields, this value must
    * be zero. For extent fields, this value must be a positive number less than or equal to the
    * size of the extent field. Unknown value is invalid in either case.
    * 
    * @param   index
    *          Index value (unadjusted) passed to BUFFER-VALUE method.
    * 
    * @return  <code>true</code> if <code>index</code> is valid, else <code>false</code>.
    * 
    * @throws  ErrorConditionException
    *          if <code>index</code> is invalid and session is not in silent error mode.
    */
   private boolean validateBufferValueIndex(NumberType index)
   {
      int max = 0;
      Method sizer = fieldRef.getSizer();
      if (sizer != null)
      {
         try
         {
            Object dmo = fieldRef.getDMO();
            integer size = (integer) sizer.invoke(dmo);
            max = size.intValue();
         }
         catch (IllegalAccessException | IllegalArgumentException | InvocationTargetException exc)
         {
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.log(Level.FINE, "Error determining DMO field size", exc);
            }
            
            return false;
         }
      }
      
      int i = (index.isUnknown() ? -1 : index.intValue());
      if (i < 0 || i > max || (max == 0 && i != 0))
      {
         String property = fieldRef.getPropertyName();
         String fieldName = TableMapper.getLegacyFieldName(getDMOProxy(), property);
         String msg = String.format("%s BUFFER-VALUE argument must be valid array index, less " +
                                    "than or equal to field extent %d",
                                    fieldName,
                                    max);
         ErrorManager.recordOrThrowError(7365, msg, false);
         
         return false;
      }
      
      return true;
   }
}
